import pytest

from lokii.model.gen_module import GenRun, GenNodeModule as Gm
from lokii.parse.node_parser import NodeParser

# noinspection SpellCheckingInspection
pytestmark = [pytest.mark.usefixtures("found_mods", "loaded_mods")]


# noinspection SpellCheckingInspection, PyTestParametrized
@pytest.mark.parametrize("found_mods, loaded_mods", [([], [])], indirect=True)
def test_parse_should_log_warning_if_no_modules_found(caplog):
    NodeParser("/test/path").parse()
    assert "No generation node file found" in caplog.text


# noinspection SpellCheckingInspection, PyTestParametrized, PyTypeChecker
@pytest.mark.parametrize(
    "found_mods, loaded_mods", [(["t1"], [Gm([{}], None)])], indirect=True
)
def test_parse_should_use_filename_if_name_not_provided(found_mods):
    parsed = NodeParser("/test/path").parse()
    for module_path in found_mods:
        node_name = module_path.replace(".gen.py", "")
        run_key = GenRun.create_key(node_name, 0)
        # expect provided filename is used for parsed gen runs
        assert node_name == parsed[run_key].node_name
        assert run_key == parsed[run_key].run_key


# noinspection SpellCheckingInspection, PyTestParametrized, PyTypeChecker
@pytest.mark.parametrize(
    "found_mods, loaded_mods", [(["t1"], [Gm([{}], name="pt1")])], indirect=True
)
def test_parse_should_use_name_instead_of_filename(found_mods, loaded_mods):
    parsed = NodeParser("/test/path").parse()
    for i, module_path in enumerate(found_mods):
        mod = loaded_mods[i]
        run_key = GenRun.create_key(mod.name, 0)
        # expect provided module name is used for parsed gen runs
        assert mod.name == parsed[run_key].node_name
        assert run_key == parsed[run_key].run_key


# noinspection SpellCheckingInspection, PyTestParametrized, PyTypeChecker
@pytest.mark.parametrize(
    "found_mods, loaded_mods", [(["t1"], [Gm([{}], version="pv1")])], indirect=True
)
def test_parse_should_use_version_instead_of_file_hash(found_mods, loaded_mods):
    parsed = NodeParser("/test/path").parse()
    for i, module_path in enumerate(found_mods):
        mod = loaded_mods[i]
        node_name = module_path.replace(".gen.py", "")
        run_key = GenRun.create_key(node_name, 0)
        # expect provided module version is used for parsed gen runs
        assert mod.version == parsed[run_key].node_version


# noinspection SpellCheckingInspection, PyTestParametrized, PyTypeChecker
@pytest.mark.parametrize(
    "found_mods, loaded_mods", [(["t1"], [Gm([{}, {}])])], indirect=True
)
def test_parse_should_wait_its_own_previous_runs(found_mods, loaded_mods):
    parsed = NodeParser("/test/path").parse()
    for i, module_path in enumerate(found_mods):
        mod = loaded_mods[i]
        node_name = module_path.replace(".gen.py", "")
        for j, r in enumerate(mod.runs):
            if j > 0:
                before_run_key = GenRun.create_key(node_name, j - 1)
                run_key = GenRun.create_key(node_name, j)
                assert before_run_key in parsed[run_key].wait


# noinspection SpellCheckingInspection, PyTestParametrized, PyTypeChecker
@pytest.mark.parametrize("found_mods", [["t1"]], indirect=True)
@pytest.mark.parametrize(
    "loaded_mods, expect",
    [
        ([Gm(None)], "`runs` configuration not found"),
        ([Gm("None")], "`runs` must be list"),
        ([Gm([])], "`runs` has no items"),
        ([Gm([{"source": []}])], "[`source`] must be str"),
        ([Gm([{"wait": ""}])], "[`wait`] must be list"),
        ([Gm([{"func": None}])], "[`func`] not found"),
        ([Gm([{"func": ""}])], "[`func`] must be function"),
        ([Gm([{"func": lambda x, y: x}])], "[`func`] must accept only one"),
    ],
    indirect=["loaded_mods"],
)
def test_parse_raise_error_if_module_not_valid(expect):
    with pytest.raises(AssertionError) as err:
        NodeParser("/test/path").parse()
    assert expect in str(err.value)



import argparse
import logging
import sys

from pathlib import Path
from typing import Optional

from lokii import Lokii
from lokii.logger.context import LoggingContext
from lokii.config import CONFIG

LOKII_ASCII = r"""
 __         ______     __  __     __     __    
/\ \       /\  __ \   /\ \/ /    /\ \   /\ \   
\ \ \____  \ \ \/\ \  \ \  _"-.  \ \ \  \ \ \  
 \ \_____\  \ \_____\  \ \_\ \_\  \ \_\  \ \_\ 
  \/_____/   \/_____/   \/_/\/_/   \/_/   \/_/ 
"""
LOKII_EPILOG = f""""""


class Command:
    def __init__(self, argv: Optional[str] = None) -> None:
        self.argv = argv.split() if argv else sys.argv[:]
        self.prog_name = Path(self.argv[0]).name

    def execute(self) -> None:
        """
        Given the command-line arguments, this creates a parser appropriate
        to that command, and runs it.
        """

        formatter_class = argparse.RawDescriptionHelpFormatter
        parser = argparse.ArgumentParser(
            prog=self.prog_name,
            description=f"{LOKII_ASCII}\n{self.prog_name} version {CONFIG.version}",
            epilog=LOKII_EPILOG,
            formatter_class=formatter_class,
        )

        parser.add_argument(
            "--version", action="version", version=f"%(prog)s {CONFIG.version}"
        )

        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="set logging level to DEBUG that shows all events",
        )

        parser.add_argument(
            "-q",
            "--quiet",
            action="store_true",
            help="set logging level to CRITICAL that shows only blocking events",
        )

        parser.add_argument(
            "-l",
            "--log-file",
            action="store",
            help="set the file path where the log file will be created",
        )

        parser.add_argument(
            "-f",
            "--source-folder",
            action="store",
            metavar="PATH",
            default=".",
            type=str,
            help="generate the specified number of outputs",
        )

        parser.add_argument(
            "-o",
            "--out-folder",
            action="store",
            metavar="PATH",
            default="./data",
            type=str,
            help="redirect output to a file",
        )

        parser.add_argument(
            "-p",
            "--purge",
            action="store_true",
            help="purge cache database after generation",
        )

        arguments = parser.parse_args(self.argv[1:])

        verbose = logging.INFO
        if arguments.verbose:
            verbose = logging.DEBUG
        if arguments.quiet:
            verbose = logging.ERROR

        with LoggingContext(level=verbose, filename=arguments.log_file):
            try:
                print(LOKII_ASCII)
                lokii_gen = Lokii(arguments.source_folder, arguments.out_folder)
                lokii_gen.generate(arguments.purge)
            except Exception as err:
                logging.critical(str(err), exc_info=True)


def exec_cmd(argv: Optional[str] = None) -> None:
    """A simple method that runs a Command."""
    command = Command(argv)
    command.execute()



import pytest
from unittest.mock import Mock

from lokii import Lokii


@pytest.fixture
def found_mods(mocker, request):
    gen_files = request.param or []
    gen_files = [m if ".gen.py" in m else f"{m}.gen.py" for m in gen_files]
    mocker.patch("glob.glob", return_value=gen_files)
    return gen_files


@pytest.fixture
def loaded_mods(mocker, found_mods, request):
    def module_file_loader_side_effect(file_path: str):
        mods = request.param or []
        mod_i = found_mods.index(file_path)
        if mods[mod_i].runs and isinstance(mods[mod_i].runs, list):
            mods[mod_i].runs = [
                {"source": "SELECT * FROM range(100)", "func": lambda x: x, **run}
                for run in mods[mod_i].runs
            ]
        loader = {
            "load": Mock(),
            "module": mods[mod_i],
            "version": mods[mod_i].version or "v1",
        }
        return type("ModuleFileLoader", (object,), loader or {})()

    mock = mocker.patch("lokii.parse.node_parser.ModuleFileLoader")
    mock.side_effect = module_file_loader_side_effect
    return request.param or []


@pytest.fixture
def mock_module_loader(mocker):
    def patch(files: list[str] = None, node: dict = None) -> None:
        mocker.patch("glob.glob", return_value=files or [])
        mocker.patch("inspect.getsource", return_value="string code content")
        m = type("GenNodeModule", (object,), node or {})()
        mock = mocker.patch("lokii.parse.node_parser.ModuleFileLoader")
        mock.return_value.module = m
        mock.return_value.version = m.version if hasattr(m, "version") else "v1"

    return patch


@pytest.fixture(scope="function")
def setup_test_env(request):
    def teardown():
        Lokii.clean_env(force=True)

    Lokii.setup_env()
    request.addfinalizer(teardown)



import networkx as nx

from lokii.model.gen_module import GenRun


class GraphAnalyzer:
    def __init__(self, gen_runs: list[GenRun]):
        """
        Reads and validates dataset configuration from filesystem structure.

        :param gen_runs: root path of the dataset generation schema
        """
        self.gen_runs = gen_runs

        # initialize directed graph
        self.run_dig = nx.DiGraph()
        for run in self.gen_runs:
            if not run.wait:
                self.run_dig.add_node(run.run_key)
            for dep in run.wait:
                self.run_dig.add_edge(dep, run.run_key)

    def dependencies(self, run_key: str) -> list[str]:
        return list(nx.ancestors(self.run_dig, run_key))

    def execution_order(self) -> list[str]:
        cycles = list(nx.simple_cycles(self.run_dig))
        if len(cycles) != 0:
            raise AssertionError(f"Found cyclic dependencies!\n{cycles}")
        return list(nx.topological_sort(self.run_dig))



from __future__ import annotations

from typing import TypedDict, Callable
from typing_extensions import NotRequired

GenFunc = Callable[[dict], dict]


class GenRunConf(TypedDict):
    """
    Generation run configuration that defines how dataset should be generated.
    :var source: source SQL statement
    :var wait: list of dependent runs
    :var func: generation function
    """

    source: str
    wait: NotRequired[list[str]]
    func: Callable


class GenNodeModule:
    name: str | None
    version: str | None
    runs: list[GenRunConf]

    def __init__(self, runs: list[GenRunConf], name: str = None, version: str = None):
        self.runs = runs
        self.name = name
        self.version = version


class GenRun:
    def __init__(self, node_name: str, version: str, index: int, run_conf: GenRunConf):
        self.node_name = node_name
        self.node_version = version
        self.run_order = index
        self.run_key = GenRun.create_key(node_name, index)

        self.source = run_conf["source"]
        self.func = run_conf["func"]
        self.wait = []
        if index > 0:
            self.wait.append(GenRun.create_key(node_name, index - 1))
        for dep in run_conf["wait"]:
            self.wait.append(dep if GenRun.is_key(dep) else GenRun.create_key(dep, 0))

    @staticmethod
    def create_key(node_name: str, index: int) -> str:
        return "/".join([node_name, str(index)])

    @staticmethod
    def is_key(run_key: str) -> bool:
        return "/" in run_key



from io import StringIO
from tqdm import tqdm


class ProgressLogger(StringIO):
    def __init__(self, total: float):
        super(ProgressLogger, self).__init__()
        self.total = total
        fmt = " ".join(
            [
                "{bar}",
                "{percentage:3.0f}%",
                "{n_fmt}/{total_fmt}",
                "[{elapsed}<{remaining}, {rate_fmt}{postfix}]",
            ]
        )
        self.pbar = tqdm(total=total, bar_format=fmt, unit_scale=True, leave=False)

    def update(self, count: float):
        self.pbar.update(count)
        if count / self.total == 1:
            self.pbar.close()



from __future__ import annotations

import os
import sys
import hashlib
import inspect
from types import ModuleType
from importlib.util import spec_from_file_location, module_from_spec


class ModuleFileLoader:
    def __init__(self, file_path: str):
        """
        Loads module from given path.

        :param file_path: file path of the module
        """
        self.path = os.path.abspath(file_path)
        self.filename = os.path.basename(file_path)

        self.module: ModuleType | None = None
        self.version: str | None = None

    def load(self) -> None:
        if not os.path.exists(self.path):
            raise FileNotFoundError(f"No module file found at: {self.path}")

        # read file spec from given path. filename is used as module name
        spec = spec_from_file_location(self.filename, self.path)
        module = module_from_spec(spec)

        # introduce module to system
        sys.modules[self.filename] = module
        spec.loader.exec_module(module)
        self.module = module

        # acquire module source code as str
        src = inspect.getsource(module).encode("utf-8")
        h = hashlib.blake2b(digest_size=20)
        h.update(src)

        # hash source code to detect changes
        self.version = h.hexdigest()



import os.path

import duckdb
import pytest

from lokii.config import CONFIG
from lokii.storage.data_storage import DataStorage
from lokii.storage.temp_storage import TempStorage


@pytest.mark.usefixtures("setup_test_env")
def test_init_should_create_database_and_meta_table():
    storage = DataStorage()
    with storage.connect() as conn:
        assert os.path.exists(CONFIG.temp.db_path)
        assert "__meta" in conn.execute("SHOW TABLES;").fetchone()


@pytest.mark.usefixtures("setup_test_env")
def test_count_should_return_row_count_for_query_result():
    storage = DataStorage()
    assert storage.count("SELECT unnest([1, 2, 3])") == 3


@pytest.mark.usefixtures("setup_test_env")
def test_exec_should_return_query_result_for_given_page():
    storage = DataStorage()
    q = "SELECT * FROM range(0, 25)"
    first_page = storage.exec(q, 0, 20)
    second_page = storage.exec(q, 1, 20)
    assert len(first_page) == 20
    assert first_page[0]["range"] == 0
    assert len(second_page) == 5
    assert second_page[0]["range"] == 20


@pytest.mark.usefixtures("setup_test_env")
def test_save_should_insert_record_to_meta_table():
    storage = DataStorage()
    with storage.connect() as conn:
        storage.save("test_gen1", "test_key1", "test_v1")
        records = conn.execute("SELECT * FROM main.__meta").fetchall()
        assert len([r for r in records if "test_key1" in r]) == 1


@pytest.mark.usefixtures("setup_test_env")
def test_save_should_replace_record_with_same_key():
    storage = DataStorage()
    with storage.connect() as conn:
        storage.save("test_gen1", "test_key1", "test_v1")
        storage.save("test_gen2", "test_key1", "test_v4")
        records = conn.execute("SELECT * FROM main.__meta").fetchall()
        assert len([r for r in records if "test_key1" in r]) == 1


@pytest.mark.usefixtures("setup_test_env")
def test_meta_should_return_data_for_given_keys():
    storage = DataStorage()
    storage.save("test_gen1", "test_key1", "test_v1")
    storage.save("test_gen1", "test_key2", "test_v1")
    storage.save("test_gen2", "test_key3", "test_v4")
    deps = storage.meta(["test_key2", "test_key3"])
    assert len(deps) == 2
    assert len([d for d in deps if d["run_key"] == "test_key2"]) == 1
    assert len([d for d in deps if d["run_key"] == "test_key3"]) == 1


@pytest.mark.usefixtures("setup_test_env")
@pytest.mark.parametrize("node, expect", [("s1.n2", "s1"), ("s2.n2", "s2")])
def test_insert_should_create_schema_if_node_name_contain_dot(node, expect):
    storage = DataStorage()
    with storage.connect() as conn:
        _temp = TempStorage("_")
        _temp.dump([{"data": i} for i in range(10)])
        storage.insert(node, _temp.batches)
        res = conn.execute(
            f"SELECT * FROM information_schema.schemata WHERE schema_name = '{expect}';"
        ).fetchone()
        assert expect in res


@pytest.mark.usefixtures("setup_test_env")
@pytest.mark.parametrize("loop, expect", [([7], 7), ([12, 10, 7], 29)])
def test_insert_should_create_table_from_generated_files(loop, expect):
    storage = DataStorage()
    with storage.connect() as conn:
        _temp = TempStorage("_")
        for count in loop:
            _temp.dump([{"data": i} for i in range(count)])
        storage.insert("n1", _temp.batches)
        assert expect in conn.execute(f"SELECT COUNT() FROM n1;").fetchone()


@pytest.mark.usefixtures("setup_test_env")
@pytest.mark.parametrize("nodes, fmt", [(["n1", "n2"], "csv")])
def test_export_should_create_export_data_files(nodes, fmt):
    _temp = TempStorage("_")
    _temp.dump([{"data": i} for i in range(100)])
    storage = DataStorage()
    for node in nodes:
        storage.insert(node, _temp.batches)
    storage.export("data", fmt)
    for node in nodes:
        assert os.path.exists(os.path.join("data", f"{node}.{fmt}"))



import logging
import os
import shutil
import uuid

from lokii.config import CONFIG
from lokii.model.gen_module import GenRun
from lokii.parse.node_parser import NodeParser
from lokii.parse.graph_analyzer import GraphAnalyzer
from lokii.storage.data_storage import DataStorage
from lokii.exec.gen_executor import GenExecutor
from lokii.util.perf_timer_context import PerfTimerContext


class Lokii:
    def __init__(self, source_folder: str, out_folder: str):
        """
        Generates massive amount of relational mock data.

        :param source_folder: path of root folder that contains schema and table definitions
        :param out_folder: path of output folder for mock data
        """
        self.__source_folder = source_folder
        self.__out_folder = out_folder
        self.__gen_id = str(uuid.uuid4())

        Lokii.setup_env()
        self.__data_storage = DataStorage()
        self.__gen_parser = NodeParser(self.__source_folder)

    def generate(self, purge: bool = False):
        with PerfTimerContext() as t:
            gen_runs = self.__gen_parser.parse()
            analyzer = GraphAnalyzer(list(gen_runs.values()))
            run_exec_order = analyzer.execution_order()

            total_target_count, total_item_count = 0, 0
            for run_key in run_exec_order:
                run = gen_runs[run_key]
                dep_keys = analyzer.dependencies(run_key)
                if self.is_dataset_valid(run, dep_keys):
                    logging.info(f"{run_key} not changed. Using existing dataset.")
                    continue

                # generate dataset
                target_count, item_count, file_paths = self.generate_dataset(run)
                total_target_count += target_count
                total_item_count += item_count

                # save generation metadata in database
                self.__data_storage.save(self.__gen_id, run.run_key, run.node_version)
                # insert generated data in database
                self.__data_storage.insert(run.node_name, file_paths)

        logging.info("Generation completed!")
        logging.info("Total target item count: {:,}".format(total_target_count))
        logging.info("Generated {:,} items in {}".format(total_item_count, t))
        self.__data_storage.export(self.__out_folder, "csv")
        Lokii.clean_env(purge)

    def generate_dataset(self, gen_run: GenRun) -> (int, int, list[str]):
        with PerfTimerContext() as t:
            logger = logging.getLogger(gen_run.run_key)

            executor = GenExecutor(gen_run, self.__data_storage)
            target_count = executor.prepare_node()

            logger.info("Generation started for target {:,} items".format(target_count))
            generated_file_paths = executor.exec_node()
            item_count = executor.g_count

        logger.info("{:,} items generated in {}".format(item_count, t))
        return target_count, item_count, generated_file_paths

    def is_dataset_valid(self, run: GenRun, dep_keys: list[str]):
        metadata = self.__data_storage.meta([*dep_keys, run.run_key])
        curr = [m for m in metadata if m["run_key"] == run.run_key]
        if len(curr) == 0:
            # no dataset generated before with this run key
            return False  # must generate
        if curr[0]["version"] != run.node_version:
            # code version is different from previous run
            return False  # must regenerate
        for dep_meta in metadata:
            if dep_meta["gen_id"] == self.__gen_id:
                # dependent dataset changed in current generation
                return False  # must regenerate
        # no code changes, no dependencies changed
        return True  # dataset is valid, do not regenerate

    @staticmethod
    def setup_env():
        Lokii.clean_env(False)
        if not os.path.exists(CONFIG.temp.data_path):
            os.makedirs(CONFIG.temp.data_path)

    @staticmethod
    def clean_env(force: bool = False):
        if force and os.path.exists(CONFIG.temp.dir_path):
            shutil.rmtree(CONFIG.temp.dir_path)
        elif os.path.exists(CONFIG.temp.data_path):
            shutil.rmtree(CONFIG.temp.data_path)



import os
from pathlib import Path

root = Path(__file__).resolve().parent.parent
# read app version from root of the project
VERSION = (root / "VERSION").read_text(encoding="utf-8").strip()

# temp directory that contains generated runtime files and database
TEMP_DIR_PATH = os.environ.get("LOKII__TEMP_DIR_PATH", ".temp")
# duckdb database that stores generated data in relational tables
TEMP_DB_FILE = os.environ.get("LOKII__TEMP_DB_FILE", "lokii.duckdb")
# name of the temp data file that contains generated runtime files
TEMP_DATA_DIR = os.environ.get("LOKII__TEMP_DATA_DIR", "data")

# file extension to look for when finding generation config files
GEN_FILE_EXT = os.environ.get("LOKII__GEN_FILE_EXT", ".gen.py")
# -
GEN_CONCURRENCY = os.environ.get("LOKII__GEN_CONCURRENCY", os.cpu_count())
# -
GEN_BATCH_SIZE = os.environ.get("LOKII__GEN_BATCH_SIZE", 100_000)
# -
GEN_CHUNK_SIZE = os.environ.get("LOKII__GEN_CHUNK_SIZE", 200)


class __Config:
    class __TempConfig:
        """
        Global configuration for storing temp file paths.
        """

        @property
        def dir_path(self) -> str:
            return TEMP_DIR_PATH

        @property
        def db_path(self) -> str:
            return os.path.join(TEMP_DIR_PATH, TEMP_DB_FILE)

        @property
        def data_path(self) -> str:
            return os.path.join(TEMP_DIR_PATH, TEMP_DATA_DIR)

    class __GenConfig:
        """
        Global configuration for storing generation information.
        """

        @property
        def file_ext(self) -> str:
            return GEN_FILE_EXT

        @property
        def concurrency(self) -> int:
            return int(GEN_CONCURRENCY)

        @property
        def batch_size(self) -> int:
            return int(GEN_BATCH_SIZE)

        @property
        def chunk_size(self) -> int:
            return int(GEN_CHUNK_SIZE)

    @property
    def version(self):
        return VERSION

    @property
    def temp(self):
        return self.__TempConfig()

    @property
    def gen(self):
        return self.__GenConfig()


CONFIG = __Config()



from logging import Formatter, LogRecord

from lokii.logger.color import default_formats


class MultiFormatter(Formatter):
    """Format log messages differently for each log level"""

    def __init__(self, formats: dict[int, str] = None, **kwargs):
        base_format = kwargs.pop("fmt", None)
        super().__init__(base_format, **kwargs)
        fmts = formats or default_formats
        self.formatters = {lvl: Formatter(fmt, **kwargs) for lvl, fmt in fmts.items()}

    def format(self, record: LogRecord):
        formatter = self.formatters.get(record.levelno)
        return super().format(record) if formatter is None else formatter.format(record)



import os
import json

from lokii.config import CONFIG


class TempStorage:
    def __init__(self, node_name: str):
        """
        Temporary filesystem storage implementation for storing data generated between batches.
        It only stores data temporary and deletes all files after node generation completed.
        :param node_name: name of the node
        """
        self.node_name = node_name
        self.batches: list[str] = []
        self.item_count = 0

    def dump(self, batch_data: list[dict]) -> None:
        storage_key = self.node_name + str(len(self.batches))
        storage_path = os.path.join(CONFIG.temp.data_path, f"{storage_key}.json")

        with open(storage_path, "w") as _f:
            _f.write(json.dumps(batch_data))

        self.batches.append(storage_path)
        self.item_count = len(batch_data)



import time

from lokii.util.perf_timer_context import PerfTimerContext


def test_timer_should_track_elapsed_time():
    with PerfTimerContext() as t:
        time.sleep(0.1)
    assert 0.1 == round(t.time, 1)


def test_timer_should_track_only_in_context():
    with PerfTimerContext() as t:
        time.sleep(0.1)
    time.sleep(0.2)
    assert 0.1 == round(t.time, 1)


def test_timer_print_two_digits_over_seconds():
    t = PerfTimerContext()
    t.time = 1.1234
    assert "1.12s" == str(t)


def test_timer_print_four_digits_under_seconds():
    t = PerfTimerContext()
    t.time = 0.1234
    assert "0.1234s" == str(t)



import pytest

from model.gen_module import GenRun, GenRunConf
from lokii.parse.graph_analyzer import GraphAnalyzer

conf: GenRunConf = {"source": "SELECT 1", "wait": [], "func": lambda x: x}


# noinspection PyTypeChecker
def test_dependencies_should_return_correct_run_ids():
    r1, r2, r3 = (
        GenRun("n1", "v1", 0, {**conf, "wait": ["n3"]}),
        GenRun("n2", "v1", 0, {**conf, "wait": []}),
        GenRun("n3", "v1", 0, {**conf, "wait": ["n2"]}),
    )
    analyzer = GraphAnalyzer([r1, r2, r3])
    deps = analyzer.dependencies(r1.run_key)
    assert r3.run_key in deps
    assert r2.run_key in deps


# noinspection PyTypeChecker
def test_execution_order_should_raise_error_if_cyclic_dependencies_exists():
    r1, r2 = (
        GenRun("n1", "v1", 0, {**conf, "wait": ["n2"]}),
        GenRun("n2", "v1", 0, {**conf, "wait": ["n1"]}),
    )
    analyzer = GraphAnalyzer([r1, r2])
    with pytest.raises(AssertionError) as err:
        analyzer.execution_order()
    assert "Found cyclic dependencies" in str(err.value)


# noinspection PyTypeChecker
def test_execution_order_should_return_run_ids_in_correct_order():
    r1, r2, r3 = (
        GenRun("n1", "v1", 0, {**conf, "wait": ["n3"]}),
        GenRun("n2", "v1", 0, {**conf, "wait": ["n1", "n3"]}),
        GenRun("n3", "v1", 0, {**conf, "wait": []}),
    )
    analyzer = GraphAnalyzer([r1, r2, r3])
    order = analyzer.execution_order()
    assert order == [r3.run_key, r1.run_key, r2.run_key]



from __future__ import annotations

import logging
import glob
import inspect
from os import path

from lokii.config import CONFIG
from lokii.model.gen_module import GenNodeModule, GenRunConf, GenRun
from lokii.util.module_file_loader import ModuleFileLoader
from lokii.util.perf_timer_context import PerfTimerContext


class NodeParser:
    def __init__(self, source_folder: str):
        """
        Reads and validates dataset configuration from filesystem structure.

        :param source_folder: root path of the dataset generation schema
        """
        self.root = source_folder
        self.gen_runs: dict[str, GenRun] = {}

    def parse(self) -> dict[str, GenRun]:
        with PerfTimerContext() as t:
            gen_nodes = list(self.__parse_nodes())
            self.gen_runs = {
                GenRun.create_key(n.name, i): GenRun(n.name, n.version, i, r)
                for n in gen_nodes
                for i, r in enumerate(n.runs)
            }

        node_count = len(gen_nodes)
        run_count = len(self.gen_runs)
        if node_count == 0:
            logging.warning(f"No generation node file found at {self.root}")
        else:
            logging.info("Total {} runs found.".format(run_count))
            logging.info("{} generation node parsed in {}".format(node_count, t))
        return self.gen_runs

    def __parse_nodes(self) -> list[GenNodeModule] | None:
        glob_path = path.join(self.root, f"**/*{CONFIG.gen.file_ext}")
        file_paths = [f for f in glob.glob(glob_path)]

        for fp in file_paths:
            loader = ModuleFileLoader(fp)
            loader.load()
            m = loader.module

            assert hasattr(m, "runs"), f"`runs` configuration not found at {fp}"
            assert m.runs is not None, f"`runs` configuration not found at {fp}"
            assert isinstance(m.runs, list), f"`runs` must be list at {fp}"
            assert len(m.runs) > 0, f"`runs` has no items at {fp}"

            parsed = GenNodeModule(
                loader.module.runs,
                path.basename(fp).replace(CONFIG.gen.file_ext, ""),
                loader.version,
            )
            if hasattr(m, "name") and m.name is not None:
                parsed.name = m.name
            if hasattr(m, "version") and m.version is not None:
                parsed.version = m.version

            runs: list[GenRunConf] = []
            for i, r in enumerate(m.runs):
                runs.append(self.__parse_run(i, r, fp))

            parsed.runs = runs
            yield parsed

    def __parse_run(self, i: int, r: GenRunConf, fp: str) -> GenRunConf:
        assert "source" in r, f"runs[{i}][`source`] not found at {fp}"
        if not isinstance(r["source"], str):
            raise AssertionError(f"runs[{i}][`source`] must be str at {fp}")
        if "wait" not in r:
            r["wait"] = []
        elif not isinstance(r["wait"], list):
            raise AssertionError(f"runs[{i}][`wait`] must be list at {fp}")
        if "func" not in r or r["func"] is None:
            raise AssertionError(f"runs[{i}][`func`] not found at {fp}")
        elif not inspect.isfunction(r["func"]):
            raise AssertionError(f"runs[{i}][`func`] must be function at {fp}")
        elif len(inspect.signature(r["func"]).parameters) != 1:
            raise AssertionError(
                f"runs[{i}][`func`] must accept only one parameter at {fp}"
            )
        return r



from lokii.storage.temp_storage import TempStorage


def test_dump_should_save_to_temp_file_path_dir(mocker):
    mocker.patch("builtins.open")
    mocker.patch("os.path.exists")
    mock = mocker.patch("json.dumps")

    data = [{"data": "test1"}, {"data": "test2"}]
    TempStorage("test").dump(data)
    assert mock.call_count
    assert data in mock.call_args.args



import shutil

import pytest
import os.path
from io import StringIO

from config import CONFIG
from lokii import Lokii
from lokii.cli import exec_cmd, LOKII_ASCII
from lokii.model.gen_module import GenNodeModule as Gm


def test_exec_cmd_should_print_help(mocker):
    mock: StringIO = mocker.patch("sys.stdout", new=StringIO())
    with pytest.raises(SystemExit):
        exec_cmd("lokii -h")
    assert LOKII_ASCII in mock.getvalue()
    Lokii.clean_env(force=True)


def test_exec_cmd_should_create_log_file():
    log_file_path = "test.log"
    exec_cmd(f"lokii -l {log_file_path}")
    assert os.path.exists(log_file_path)
    os.remove(log_file_path)


def test_exec_cmd_should_obey_log_level(caplog):
    exec_cmd("lokii -q")
    assert len(caplog.text) == 0
    exec_cmd("lokii -v")
    assert len(caplog.text) > 0


# noinspection SpellCheckingInspection, PyTestParametrized, PyTypeChecker
@pytest.mark.usefixtures("found_mods", "loaded_mods")
@pytest.mark.parametrize(
    "found_mods, loaded_mods",
    [(["t1"], [Gm([{"source": "INVALID SQL"}])])],
    indirect=True,
)
def test_exec_cmd_should_log_errors_with_critical_level(caplog):
    exec_cmd("lokii -f .")
    assert "CRITICAL" in caplog.text


# noinspection SpellCheckingInspection, PyTestParametrized, PyTypeChecker
@pytest.mark.usefixtures("found_mods", "loaded_mods")
@pytest.mark.parametrize("found_mods, loaded_mods", [([], [])], indirect=True)
def test_exec_cmd_should_clear_db_if_purge():
    exec_cmd("lokii -p")
    assert not os.path.exists(CONFIG.temp.db_path)
    assert os.path.exists("data")
    shutil.rmtree("data")


# noinspection SpellCheckingInspection, PyTestParametrized, PyTypeChecker
@pytest.mark.usefixtures("found_mods", "loaded_mods")
@pytest.mark.parametrize("outpath", ["test_out_1"])
@pytest.mark.parametrize(
    "found_mods, loaded_mods",
    [(["t1.gen.py"], [Gm([{"source": "SELECT * FROM range(100)"}])])],
    indirect=True,
)
def test_exec_cmd_should_execute_generation(outpath):
    exec_cmd(f"lokii -p -o {outpath} -f .")
    assert os.path.exists(outpath)
    shutil.rmtree(outpath)


# noinspection SpellCheckingInspection, PyTestParametrized, PyTypeChecker
@pytest.mark.usefixtures("found_mods", "loaded_mods")
@pytest.mark.parametrize("outpath", ["test_out_1"])
@pytest.mark.parametrize(
    "found_mods, loaded_mods",
    [(["t1.gen.py"], [Gm([{"source": "SELECT * FROM range(100)"}])])],
    indirect=True,
)
def test_exec_cmd_should_cache_consequent_runs(caplog, outpath):
    exec_cmd(f"lokii -o {outpath} -f .")
    assert os.path.exists(CONFIG.temp.db_path)
    caplog.clear()

    # start generation for second time without any code or dependency change
    exec_cmd(f"lokii -p -o {outpath} -f .")
    assert "not changed. Using existing dataset." in caplog.text
    assert os.path.exists(outpath)
    shutil.rmtree(outpath)


# noinspection PyTypeChecker
gen_mod = Gm([{"source": "SELECT * FROM range(100)"}], version="v1")


# noinspection SpellCheckingInspection, PyTestParametrized
@pytest.mark.usefixtures("found_mods", "loaded_mods")
@pytest.mark.parametrize("outpath", ["test_out_1"])
@pytest.mark.parametrize(
    "found_mods, loaded_mods", [(["t1.gen.py"], [gen_mod])], indirect=True
)
def test_exec_cmd_should_regenerate_on_code_change(caplog, outpath):
    # start generation for code version v1
    gen_mod.version = "v1"
    exec_cmd(f"lokii -o {outpath} -f .")
    assert os.path.exists(CONFIG.temp.db_path)
    caplog.clear()

    # start generation for code version v2
    gen_mod.version = "v2"
    exec_cmd("lokii -p -o data -f .")
    # should regenerate and not use cache
    assert "not changed. Using existing dataset." not in caplog.text
    assert os.path.exists(outpath)
    shutil.rmtree(outpath)


# noinspection PyTypeChecker
gen_mod1, gen_mod2 = (
    Gm([{"source": "SELECT * FROM range(100)", "wait": ["n2"]}], name="n1"),
    Gm([{"source": "SELECT * FROM range(100)"}], name="n2", version="v1"),
)


# noinspection SpellCheckingInspection, PyTestParametrized
@pytest.mark.usefixtures("found_mods", "loaded_mods")
@pytest.mark.parametrize("outpath", ["test_out_1"])
@pytest.mark.parametrize(
    "found_mods, loaded_mods",
    [(["t1.gen.py", "t2.gen.py"], [gen_mod1, gen_mod2])],
    indirect=True,
)
def test_exec_cmd_should_regenerate_on_dependency_change(caplog, outpath):
    # start generation for dependency code version v1
    gen_mod2.version = "v1"
    exec_cmd(f"lokii -o {outpath} -f .")
    assert os.path.exists(CONFIG.temp.db_path)
    caplog.clear()

    # start generation for dependency code version v2
    gen_mod2.version = "v2"
    exec_cmd("lokii -p -o data -f .")
    # should regenerate and not use cache because of dependency change
    assert "not changed. Using existing dataset." not in caplog.text
    assert os.path.exists(outpath)
    shutil.rmtree(outpath)



#!/usr/bin/env python

from pathlib import Path

from setuptools import setup, find_packages

root = Path(__file__).resolve().parent
VERSION = (root / "VERSION").read_text(encoding="utf-8").strip()
README = (root / "README.md").read_text(encoding="utf-8")

setup(
    name="lokii",
    version=VERSION,
    packages=find_packages(where="."),
    description="Generate, Load, Develop and Test with consistent relational datasets!",
    long_description=README,
    long_description_content_type="text/markdown",
    keywords="data generation, relational datasets, development environment, testing, database",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Testing",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Typing :: Typed",
    ],
    author="Doruk Eren Aktaş",
    author_email="dorukerenaktas@gmail.com",
    py_modules=["lokii"],
    entry_points={
        "console_scripts": ["lokii=lokii.cli:exec_cmd"],
    },
    url="https://github.com/dorukerenaktas/lokii",
    data_files=[("", ["VERSION"])],
    install_requires=[
        "pandas==2.0.1",
        "pathos==0.3.0",
        "networkx==3.1",
        "tqdm==4.65.0",
        "duckdb==0.8.0",
        "typing==3.7.4.3",
        "typing-extensions==4.6.2",
    ],
    license="MIT License",
    zip_safe=False,
)



import os.path
from functools import partial

import duckdb
from typing import TypedDict

from lokii.config import CONFIG

CONN: duckdb.DuckDBPyConnection


class NodeMetadata(TypedDict):
    run_key: str
    version: str
    gen_id: str


class DataStorage:
    def __init__(self):
        """
        Temporary filesystem storage implementation for storing data generated between batches.
        It only stores data temporary and deletes all files after
        """
        self.connect = partial(duckdb.connect, database=CONFIG.temp.db_path)
        with self.connect() as conn:
            # create node meta table to store run generation information
            q = (
                "CREATE TABLE IF NOT EXISTS main.__meta"
                "(run_key TEXT, version TEXT, gen_id TEXT, PRIMARY KEY(run_key));"
            )
            conn.execute(q).fetchall()

    def count(self, query: str) -> int:
        with self.connect() as conn:
            q = f"WITH _t AS ({query}) SELECT COUNT() FROM _t;"
            (count,) = conn.execute(q).fetchone()
            return count

    def exec(self, query: str, index: int, size: int) -> list[dict]:
        with self.connect() as conn:
            q = f"WITH _t AS ({query}) SELECT * FROM _t LIMIT {size} OFFSET {index * size};"
            data = conn.execute(q).df()
            return data.to_dict("records")

    def save(self, gen_id: str, run_key: str, version: str) -> None:
        """
        Generation id and node version will be stored in a meta table that can be used to check
        if gen run data is valid for consecutive runs.

        :param gen_id: identification of the generation process
        :param run_key: identification of the generation run
        :param version: node version of the module
        """
        with self.connect() as conn:
            q = f"""
            INSERT OR REPLACE INTO main.__meta(run_key, version, gen_id)
            VALUES ('{run_key}', '{version}', '{gen_id}');
            """
            conn.execute(q).fetchall()

    def meta(self, run_keys: list[str]) -> list[NodeMetadata]:
        """
        Fetches generation metadata information about given runs.
        :param run_keys: list of run ids
        :return: list of node meta dict
        """
        with self.connect() as conn:
            keys = ",".join([f"'{k}'" for k in run_keys])
            q = f"SELECT run_key, version, gen_id FROM main.__meta WHERE run_key IN ({keys});"
            data = conn.execute(q).df()
            return data.to_dict("records")

    def insert(self, name: str, files: list[str]) -> None:
        """
        Creates a table for given node name in local relational database. If there is a table
        with the same node name it will drop all data and create a fresh one. Given files will
        be concatenated and inserted in the fresh table.

        :param name: node name of the module
        :param files: list of generated file paths
        """
        with self.connect() as conn:
            if "." in name:
                # create schema if node name defines `schema.table` syntax
                assert len(name.split(".")) == 2, "Nested schemas are not supported."
                schema = name.split(".")[0]
                q = f"CREATE SCHEMA IF NOT EXISTS {schema};"
                conn.execute(q).fetchall()

            # concatenate and insert file contents in a fresh table
            q = f"CREATE OR REPLACE TABLE {name} AS SELECT * FROM read_json_auto({files});"
            conn.execute(q).fetchall()

    def export(self, out_path: str, fmt: str) -> None:
        """
        Exports all generated tables to given file format.

        :param out_path: folder path of the exported files
        :param fmt: file format of the exported files
        """
        with self.connect() as conn:
            tables = conn.execute("SHOW TABLES;").fetchall()
            # exclude meta table from export list
            tables = [f for (f,) in tables if "__meta" != f]

            # create out folder if not exists
            if not os.path.exists(out_path):
                os.makedirs(out_path)

            for t in tables:
                path = os.path.join(out_path, f"{t}.{fmt}")
                # export to output path
                conn.execute(f"COPY {t} TO '{path}'").fetchone()
